@@ -14,7 +14,6 @@
 #include <assert.h>
 #include <ctype.h>
 #include <errno.h>
-#include <fcntl.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -36,7 +35,7 @@
 #include <haproxy/check.h>
 #include <haproxy/chunk.h>
 #include <haproxy/dgram.h>
-#include <haproxy/dynbuf-t.h>
+#include <haproxy/dynbuf.h>
 #include <haproxy/extcheck.h>
 #include <haproxy/fd.h>
 #include <haproxy/global.h>
@@ -59,7 +58,6 @@
 #include <haproxy/server.h>
 #include <haproxy/ssl_sock.h>
 #include <haproxy/stats-t.h>
-#include <haproxy/stream_interface.h>
 #include <haproxy/task.h>
 #include <haproxy/tcpcheck.h>
 #include <haproxy/thread.h>
@@ -117,7 +115,7 @@ static const struct name_desc check_trace_decoding[] = {
 #define CHK_VERB_CLEAN    1
 	{ .name="clean",    .desc="only user-friendly stuff, generally suitable for level \"user\"" },
 #define CHK_VERB_MINIMAL  2
-	{ .name="minimal",  .desc="report info on stream and stream-interfaces" },
+	{ .name="minimal",  .desc="report info on streams and connectors" },
 #define CHK_VERB_SIMPLE   3
 	{ .name="simple",   .desc="add info on request and response channels" },
 #define CHK_VERB_ADVANCED 4
@@ -142,13 +140,6 @@ struct trace_source trace_check = {
 INITCALL1(STG_REGISTER, trace_register_source, TRACE_SOURCE);
 
 
-static int wake_srv_chk(struct conn_stream *cs);
-struct data_cb check_conn_cb = {
-	.wake = wake_srv_chk,
-	.name = "CHCK",
-};
-
-
 /* Dummy frontend used to create all checks sessions. */
 struct proxy checks_fe;
 
@@ -198,14 +189,18 @@ static void check_trace(enum trace_level level, uint64_t mask,
 	if (!check || src->verbosity < CHK_VERB_CLEAN)
 		return;
 
-	chunk_appendf(&trace_buf, " : [%c] SRV=%s",
-		      ((check->type == PR_O2_EXT_CHK) ? 'E' : (check->state & CHK_ST_AGENT ? 'A' : 'H')),
-		      srv->id);
+	if (srv) {
+		chunk_appendf(&trace_buf, " : [%c] SRV=%s",
+			      ((check->type == PR_O2_EXT_CHK) ? 'E' : (check->state & CHK_ST_AGENT ? 'A' : 'H')),
+			      srv->id);
 
-	chunk_appendf(&trace_buf, " status=%d/%d %s",
-		      (check->health >= check->rise) ? check->health - check->rise + 1 : check->health,
-		      (check->health >= check->rise) ? check->fall : check->rise,
-		      (check->health >= check->rise) ? (srv->uweight ? "UP" : "DRAIN") : "DOWN");
+		chunk_appendf(&trace_buf, " status=%d/%d %s",
+			      (check->health >= check->rise) ? check->health - check->rise + 1 : check->health,
+			      (check->health >= check->rise) ? check->fall : check->rise,
+			      (check->health >= check->rise) ? (srv->uweight ? "UP" : "DRAIN") : "DOWN");
+	}
+	else
+		chunk_appendf(&trace_buf, " : [EMAIL]");
 
 	switch (check->result) {
 	case CHK_RES_NEUTRAL: res = "-";     break;
@@ -232,9 +227,11 @@ static void check_trace(enum trace_level level, uint64_t mask,
 		return;
 
 
-	if (check->cs) {
-		chunk_appendf(&trace_buf, " - conn=%p(0x%08x)", check->cs->conn, check->cs->conn->flags);
-		chunk_appendf(&trace_buf, " cs=%p(0x%08x)", check->cs, check->cs->flags);
+	if (check->sc) {
+		struct connection *conn = sc_conn(check->sc);
+
+		chunk_appendf(&trace_buf, " - conn=%p(0x%08x)", conn, conn ? conn->flags : 0);
+		chunk_appendf(&trace_buf, " sc=%p(0x%08x)", check->sc, check->sc->flags);
 	}
 
 	if (mask & CHK_EV_TCPCHK) {
@@ -368,7 +365,7 @@ static const struct analyze_status analyze_statuses[HANA_STATUS_SIZE] = {		/* 0:
  */
 static inline int unclean_errno(int err)
 {
-	if (err == EAGAIN || err == EINPROGRESS ||
+	if (err == EAGAIN || err == EWOULDBLOCK || err == EINPROGRESS ||
 	    err == EISCONN || err == EALREADY)
 		return 0;
 	return err;
@@ -431,6 +428,31 @@ const char *get_analyze_status(short analyze_status) {
 		return analyze_statuses[HANA_STATUS_UNKNOWN].desc;
 }
 
+/* append check info to buffer msg */
+void check_append_info(struct buffer *msg, struct check *check)
+{
+	if (!check)
+		return;
+	chunk_appendf(msg, ", reason: %s", get_check_status_description(check->status));
+
+	if (check->status >= HCHK_STATUS_L57DATA)
+		chunk_appendf(msg, ", code: %d", check->code);
+
+	if (check->desc[0]) {
+		struct buffer src;
+
+		chunk_appendf(msg, ", info: \"");
+
+		chunk_initlen(&src, check->desc, 0, strlen(check->desc));
+		chunk_asciiencode(msg, &src, '"');
+
+		chunk_appendf(msg, "\"");
+	}
+
+	if (check->duration >= 0)
+		chunk_appendf(msg, ", check duration: %ldms", check->duration);
+}
+
 /* Sets check->status, update check->duration and fill check->result with an
  * adequate CHK_RES_* value. The new check->health is computed based on the
  * result.
@@ -442,14 +464,14 @@ void set_server_check_status(struct check *check, short status, const char *desc
 {
 	struct server *s = check->server;
 	short prev_status = check->status;
-	int report = 0;
+	int report = (status != prev_status) ? 1 : 0;
 
 	TRACE_POINT(CHK_EV_HCHK_RUN, check);
 
 	if (status == HCHK_STATUS_START) {
 		check->result = CHK_RES_UNKNOWN;	/* no result yet */
 		check->desc[0] = '\0';
-		check->start = now;
+		check->start = now_ns;
 		return;
 	}
 
@@ -468,10 +490,10 @@ void set_server_check_status(struct check *check, short status, const char *desc
 
 	if (status == HCHK_STATUS_HANA)
 		check->duration = -1;
-	else if (!tv_iszero(&check->start)) {
+	else if (check->start) {
 		/* set_server_check_status() may be called more than once */
-		check->duration = tv_ms_elapsed(&check->start, &now);
-		tv_zero(&check->start);
+		check->duration = ns_to_ms(now_ns - check->start);
+		check->start = 0;
 	}
 
 	/* no change is expected if no state change occurred */
@@ -483,8 +505,6 @@ void set_server_check_status(struct check *check, short status, const char *desc
 	 */
 	if (!s)
 	    return;
-	report = 0;
-
 
 	switch (check->result) {
 	case CHK_RES_FAILED:
@@ -514,15 +534,17 @@ void set_server_check_status(struct check *check, short status, const char *desc
 
 		/* clear consecutive_errors if observing is enabled */
 		if (s->onerror)
-			s->consecutive_errors = 0;
+			HA_ATOMIC_STORE(&s->consecutive_errors, 0);
 		break;
 
 	default:
 		break;
 	}
 
-	if (s->proxy->options2 & PR_O2_LOGHCHKS &&
-	    (status != prev_status || report)) {
+	if (report)
+		srv_event_hdl_publish_check(s, check);
+
+	if (s->proxy->options2 & PR_O2_LOGHCHKS && report) {
 		chunk_printf(&trash,
 		             "%s check for %sserver %s/%s %s%s",
 			     (check->state & CHK_ST_AGENT) ? "Agent" : "Health",
@@ -531,7 +553,7 @@ void set_server_check_status(struct check *check, short status, const char *desc
 		             (check->result == CHK_RES_CONDPASS) ? "conditionally ":"",
 		             (check->result >= CHK_RES_PASSED)   ? "succeeded" : "failed");
 
-		srv_append_status(&trash, s, check, -1, 0);
+		check_append_info(&trash, check);
 
 		chunk_appendf(&trash, ", status: %d/%d %s",
 		             (check->health >= check->rise) ? check->health - check->rise + 1 : check->health,
@@ -544,6 +566,16 @@ void set_server_check_status(struct check *check, short status, const char *desc
 	}
 }
 
+static inline enum srv_op_st_chg_cause check_notify_cause(struct check *check)
+{
+	struct server *s = check->server;
+
+	/* We only report a cause for the check if we did not do so previously */
+	if (!s->track && !(s->proxy->options2 & PR_O2_LOGHCHKS))
+		return (check->state & CHK_ST_AGENT) ? SRV_OP_STCHGC_AGENT : SRV_OP_STCHGC_HEALTH;
+	return SRV_OP_STCHGC_NONE;
+}
+
 /* Marks the check <check>'s server down if the current check is already failed
  * and the server is not down yet nor in maintenance.
  */
@@ -564,8 +596,7 @@ void check_notify_failure(struct check *check)
 		return;
 
 	TRACE_STATE("health-check failed, set server DOWN", CHK_EV_HCHK_END|CHK_EV_HCHK_ERR, check);
-	/* We only report a reason for the check if we did not do so previously */
-	srv_set_stopped(s, NULL, (!s->track && !(s->proxy->options2 & PR_O2_LOGHCHKS)) ? check : NULL);
+	srv_set_stopped(s, check_notify_cause(check));
 }
 
 /* Marks the check <check> as valid and tries to set its server up, provided
@@ -599,7 +630,7 @@ void check_notify_success(struct check *check)
 		return;
 
 	TRACE_STATE("health-check succeeded, set server RUNNING", CHK_EV_HCHK_END|CHK_EV_HCHK_SUCC, check);
-	srv_set_running(s, NULL, (!s->track && !(s->proxy->options2 & PR_O2_LOGHCHKS)) ? check : NULL);
+	srv_set_running(s, check_notify_cause(check));
 }
 
 /* Marks the check <check> as valid and tries to set its server into stopping mode
@@ -628,7 +659,7 @@ void check_notify_stopping(struct check *check)
 		return;
 
 	TRACE_STATE("health-check condionnaly succeeded, set server STOPPING", CHK_EV_HCHK_END|CHK_EV_HCHK_SUCC, check);
-	srv_set_stopping(s, NULL, (!s->track && !(s->proxy->options2 & PR_O2_LOGHCHKS)) ? check : NULL);
+	srv_set_stopping(s, check_notify_cause(check));
 }
 
 /* note: use health_adjust() only, which first checks that the observe mode is
@@ -637,7 +668,6 @@ void check_notify_stopping(struct check *check)
 void __health_adjust(struct server *s, short status)
 {
 	int failed;
-	int expire;
 
 	if (s->observe >= HANA_OBS_SIZE)
 		return;
@@ -660,25 +690,23 @@ void __health_adjust(struct server *s, short status)
 
 	if (!failed) {
 		/* good: clear consecutive_errors */
-		s->consecutive_errors = 0;
+		HA_ATOMIC_STORE(&s->consecutive_errors, 0);
 		return;
 	}
 
-	_HA_ATOMIC_INC(&s->consecutive_errors);
-
-	if (s->consecutive_errors < s->consecutive_errors_limit)
+	if (HA_ATOMIC_ADD_FETCH(&s->consecutive_errors, 1) < s->consecutive_errors_limit)
 		return;
 
 	chunk_printf(&trash, "Detected %d consecutive errors, last one was: %s",
-	             s->consecutive_errors, get_analyze_status(status));
-
-	if (s->check.fastinter)
-		expire = tick_add(now_ms, MS_TO_TICKS(s->check.fastinter));
-	else
-		expire = TICK_ETERNITY;
+	             HA_ATOMIC_LOAD(&s->consecutive_errors), get_analyze_status(status));
 
 	HA_SPIN_LOCK(SERVER_LOCK, &s->lock);
 
+	/* force fastinter for upcoming check
+	 * (does nothing if fastinter is not enabled)
+	 */
+	s->check.state |= CHK_ST_FASTINTER;
+
 	switch (s->onerror) {
 		case HANA_ONERR_FASTINTER:
 		/* force fastinter - nothing to do here as all modes force it */
@@ -689,7 +717,7 @@ void __health_adjust(struct server *s, short status)
 			if (s->check.health > s->check.rise)
 				s->check.health = s->check.rise + 1;
 
-			/* fall through */
+			__fallthrough;
 
 		case HANA_ONERR_FAILCHK:
 		/* simulate a failed health check */
@@ -713,12 +741,15 @@ void __health_adjust(struct server *s, short status)
 
 	HA_SPIN_UNLOCK(SERVER_LOCK, &s->lock);
 
-	s->consecutive_errors = 0;
+	HA_ATOMIC_STORE(&s->consecutive_errors, 0);
 	_HA_ATOMIC_INC(&s->counters.failed_hana);
 
-	if (tick_isset(expire) && tick_is_lt(expire, s->check.task->expire)) {
-		/* requeue check task with new expire */
-		task_schedule(s->check.task, expire);
+	if (s->check.fastinter) {
+		/* timer might need to be advanced, it might also already be
+		 * running in another thread. Let's just wake the task up, it
+		 * will automatically adjust its timer.
+		 */
+		task_wakeup(s->check.task, TASK_WOKEN_MSG);
 	}
 }
 
@@ -742,6 +773,8 @@ static int retrieve_errno_from_socket(struct connection *conn)
 	if (!conn_ctrl_ready(conn))
 		return 0;
 
+	BUG_ON(conn->flags & CO_FL_FDLESS);
+
 	if (getsockopt(conn->handle.fd, SOL_SOCKET, SO_ERROR, &skerr, &lskerr) == 0)
 		errno = skerr;
 
@@ -776,8 +809,8 @@ static int retrieve_errno_from_socket(struct connection *conn)
  */
 void chk_report_conn_err(struct check *check, int errno_bck, int expired)
 {
-	struct conn_stream *cs = check->cs;
-	struct connection *conn = cs_conn(cs);
+	struct stconn *sc = check->sc;
+	struct connection *conn = sc_conn(sc);
 	const char *err_msg;
 	struct buffer *chk;
 	int step;
@@ -790,8 +823,7 @@ void chk_report_conn_err(struct check *check, int errno_bck, int expired)
 	if (conn && errno)
 		retrieve_errno_from_socket(conn);
 
-	if (conn && !(conn->flags & CO_FL_ERROR) &&
-	    !(cs->flags & CS_FL_ERROR) && !expired)
+	if (conn && !(conn->flags & CO_FL_ERROR) && !sc_ep_test(sc, SE_FL_ERROR) && !expired)
 		return;
 
 	TRACE_ENTER(CHK_EV_HCHK_END|CHK_EV_HCHK_ERR, check, 0, 0, (size_t[]){expired});
@@ -910,7 +942,7 @@ void chk_report_conn_err(struct check *check, int errno_bck, int expired)
 	}
 	else if (conn->flags & CO_FL_WAIT_L4_CONN) {
 		/* L4 not established (yet) */
-		if (conn->flags & CO_FL_ERROR || cs->flags & CS_FL_ERROR)
+		if (conn->flags & CO_FL_ERROR || sc_ep_test(sc, SE_FL_ERROR))
 			set_server_check_status(check, HCHK_STATUS_L4CON, err_msg);
 		else if (expired)
 			set_server_check_status(check, HCHK_STATUS_L4TOUT, err_msg);
@@ -925,12 +957,12 @@ void chk_report_conn_err(struct check *check, int errno_bck, int expired)
 	}
 	else if (conn->flags & CO_FL_WAIT_L6_CONN) {
 		/* L6 not established (yet) */
-		if (conn->flags & CO_FL_ERROR || cs->flags & CS_FL_ERROR)
+		if (conn->flags & CO_FL_ERROR || sc_ep_test(sc, SE_FL_ERROR))
 			set_server_check_status(check, HCHK_STATUS_L6RSP, err_msg);
 		else if (expired)
 			set_server_check_status(check, HCHK_STATUS_L6TOUT, err_msg);
 	}
-	else if (conn->flags & CO_FL_ERROR || cs->flags & CS_FL_ERROR) {
+	else if (conn->flags & CO_FL_ERROR || sc_ep_test(sc, SE_FL_ERROR)) {
 		/* I/O error after connection was established and before we could diagnose */
 		set_server_check_status(check, HCHK_STATUS_SOCKERR, err_msg);
 	}
@@ -997,9 +1029,9 @@ int httpchk_build_status_header(struct server *s, struct buffer *buf)
 		      s->queue.length);
 
 	if ((s->cur_state == SRV_ST_STARTING) &&
-	    now.tv_sec < s->last_change + s->slowstart &&
-	    now.tv_sec >= s->last_change) {
-		ratio = MAX(1, 100 * (now.tv_sec - s->last_change) / s->slowstart);
+	    ns_to_sec(now_ns) < s->last_change + s->slowstart &&
+	    ns_to_sec(now_ns) >= s->last_change) {
+		ratio = MAX(1, 100 * (ns_to_sec(now_ns) - s->last_change) / s->slowstart);
 		chunk_appendf(buf, "; throttle=%d%%", ratio);
 	}
 
@@ -1014,14 +1046,17 @@ int httpchk_build_status_header(struct server *s, struct buffer *buf)
  * It returns 0 on normal cases, <0 if at least one close() has happened on the
  * connection (eg: reconnect). It relies on tcpcheck_main().
  */
-static int wake_srv_chk(struct conn_stream *cs)
+int wake_srv_chk(struct stconn *sc)
 {
-	struct connection *conn = cs->conn;
-	struct check *check = cs->data;
+	struct connection *conn;
+	struct check *check = __sc_check(sc);
 	struct email_alertq *q = container_of(check, typeof(*q), check);
 	int ret = 0;
 
 	TRACE_ENTER(CHK_EV_HCHK_WAKE, check);
+	if (check->result != CHK_RES_UNKNOWN)
+		goto end;
+
 	if (check->server)
 		HA_SPIN_LOCK(SERVER_LOCK, &check->server->lock);
 	else
@@ -1030,10 +1065,10 @@ static int wake_srv_chk(struct conn_stream *cs)
 	/* we may have to make progress on the TCP checks */
 	ret = tcpcheck_main(check);
 
-	cs = check->cs;
-	conn = cs->conn;
+	sc = check->sc;
+	conn = sc_conn(sc);
 
-	if (unlikely(conn->flags & CO_FL_ERROR || cs->flags & CS_FL_ERROR)) {
+	if (unlikely(!conn || conn->flags & CO_FL_ERROR || sc_ep_test(sc, SE_FL_ERROR))) {
 		/* We may get error reports bypassing the I/O handlers, typically
 		 * the case when sending a pure TCP check which fails, then the I/O
 		 * handlers above are not called. This is completely handled by the
@@ -1046,20 +1081,9 @@ static int wake_srv_chk(struct conn_stream *cs)
 	}
 
 	if (check->result != CHK_RES_UNKNOWN || ret == -1) {
-		/* Check complete or aborted. If connection not yet closed do it
-		 * now and wake the check task up to be sure the result is
-		 * handled ASAP. */
-		cs_drain_and_close(cs);
+		/* Check complete or aborted. Wake the check task up to be sure
+		 * the result is handled ASAP. */
 		ret = -1;
-
-		if (check->wait_list.events)
-			cs->conn->mux->unsubscribe(cs, check->wait_list.events, &check->wait_list);
-
-		/* We may have been scheduled to run, and the
-		 * I/O handler expects to have a cs, so remove
-		 * the tasklet
-		 */
-		tasklet_remove_from_tasklet_list(check->wait_list.tasklet);
 		task_wakeup(check->task, TASK_WOKEN_IO);
 	}
 
@@ -1068,17 +1092,17 @@ static int wake_srv_chk(struct conn_stream *cs)
 	else
 		HA_SPIN_UNLOCK(EMAIL_ALERTS_LOCK, &q->lock);
 
+  end:
 	TRACE_LEAVE(CHK_EV_HCHK_WAKE, check);
 	return ret;
 }
 
 /* This function checks if any I/O is wanted, and if so, attempts to do so */
-struct task *event_srv_chk_io(struct task *t, void *ctx, unsigned int state)
+struct task *srv_chk_io_cb(struct task *t, void *ctx, unsigned int state)
 {
-	struct check *check = ctx;
-	struct conn_stream *cs = check->cs;
+	struct stconn *sc = ctx;
 
-	wake_srv_chk(cs);
+	wake_srv_chk(sc);
 	return NULL;
 }
 
@@ -1092,22 +1116,82 @@ struct task *process_chk_conn(struct task *t, void *context, unsigned int state)
 {
 	struct check *check = context;
 	struct proxy *proxy = check->proxy;
-	struct conn_stream *cs;
+	struct stconn *sc;
 	struct connection *conn;
 	int rv;
 	int expired = tick_is_expired(t->expire, now_ms);
 
 	TRACE_ENTER(CHK_EV_TASK_WAKE, check);
 
+	if (check->state & CHK_ST_SLEEPING) {
+		/* This check just restarted. It's still time to verify if
+		 * we're on an overloaded thread or if a more suitable one is
+		 * available. This helps spread the load over the available
+		 * threads, without migrating too often. For this we'll check
+		 * our load, and pick a random thread, check if it has less
+		 * than half of the current thread's load, and if so we'll
+		 * bounce the task there. It's possible because it's not yet
+		 * tied to the current thread. The other thread will not bounce
+		 * the task again because we're removing CHK_ST_SLEEPING.
+		 */
+		uint my_load = HA_ATOMIC_LOAD(&th_ctx->rq_total);
+
+		check->state &= ~CHK_ST_SLEEPING;
+
+		if (my_load >= 2) {
+			uint new_tid  = statistical_prng_range(global.nbthread);
+			uint new_load = HA_ATOMIC_LOAD(&ha_thread_ctx[new_tid].rq_total);
+
+			if (new_load <= my_load / 2) {
+				/* Found one. Let's migrate the task over there. We have to
+				 * remove it from the WQ first and kill its expire time
+				 * otherwise the scheduler will reinsert it and trigger a
+				 * BUG_ON() as we're not allowed to call task_queue() for a
+				 * foreign thread. The recipient will restore the expiration.
+				 */
+				task_unlink_wq(t);
+				t->expire = TICK_ETERNITY;
+				task_set_thread(t, new_tid);
+				task_wakeup(t, TASK_WOKEN_MSG);
+				TRACE_LEAVE(CHK_EV_TASK_WAKE, check);
+				return t;
+			}
+		}
+	}
+
 	if (check->server)
 		HA_SPIN_LOCK(SERVER_LOCK, &check->server->lock);
 
+	if (!(check->state & (CHK_ST_INPROGRESS|CHK_ST_IN_ALLOC|CHK_ST_OUT_ALLOC))) {
+		/* This task might have bounced from another overloaded thread, it
+		 * needs an expiration timer that was supposed to be now, but that
+		 * was erased during the bounce.
+		 */
+		if (!tick_isset(t->expire)) {
+			t->expire = now_ms;
+			expired = 0;
+		}
+	}
+
 	if (unlikely(check->state & CHK_ST_PURGE)) {
 		TRACE_STATE("health-check state to purge", CHK_EV_TASK_WAKE, check);
 	}
 	else if (!(check->state & (CHK_ST_INPROGRESS))) {
-		/* no check currently running */
-		if (!expired) /* woke up too early */ {
+		/* no check currently running, but we might have been woken up
+		 * before the timer's expiration to update it according to a
+		 * new state (e.g. fastinter), in which case we'll reprogram
+		 * the new timer.
+		 */
+		if (!tick_is_expired(t->expire, now_ms)) { /* woke up too early */
+			if (check->server) {
+				int new_exp = tick_add(now_ms, MS_TO_TICKS(srv_getinter(check)));
+
+				if (tick_is_expired(new_exp, t->expire)) {
+					TRACE_STATE("health-check was advanced", CHK_EV_TASK_WAKE, check);
+					goto update_timer;
+				}
+			}
+
 			TRACE_STATE("health-check wake up too early", CHK_EV_TASK_WAKE, check);
 			goto out_unlock;
 		}
@@ -1128,36 +1212,44 @@ struct task *process_chk_conn(struct task *t, void *context, unsigned int state)
 		check->state |= CHK_ST_INPROGRESS;
 		TRACE_STATE("init new health-check", CHK_EV_TASK_WAKE|CHK_EV_HCHK_START, check);
 
-		task_set_affinity(t, tid_bit);
+		task_set_thread(t, tid);
 
 		check->current_step = NULL;
+
+		check->sc = sc_new_from_check(check, SC_FL_NONE);
+		if (!check->sc) {
+			set_server_check_status(check, HCHK_STATUS_SOCKERR, NULL);
+			goto end;
+		}
 		tcpcheck_main(check);
 		expired = 0;
 	}
 
-	cs = check->cs;
-	conn = cs_conn(cs);
-
 	/* there was a test running.
 	 * First, let's check whether there was an uncaught error,
 	 * which can happen on connect timeout or error.
 	 */
 	if (check->result == CHK_RES_UNKNOWN && likely(!(check->state & CHK_ST_PURGE))) {
+		sc = check->sc;
+		conn = sc_conn(sc);
+
 		/* Here the connection must be defined. Otherwise the
 		 * error would have already been detected
 		 */
-		if ((conn && ((conn->flags & CO_FL_ERROR) || (cs->flags & CS_FL_ERROR))) || expired) {
+		if ((conn && ((conn->flags & CO_FL_ERROR) || sc_ep_test(sc, SE_FL_ERROR))) || expired) {
 			TRACE_ERROR("report connection error", CHK_EV_TASK_WAKE|CHK_EV_HCHK_END|CHK_EV_HCHK_ERR, check);
 			chk_report_conn_err(check, 0, expired);
 		}
 		else {
 			if (check->state & CHK_ST_CLOSE_CONN) {
 				TRACE_DEVEL("closing current connection", CHK_EV_TASK_WAKE|CHK_EV_HCHK_RUN, check);
-				cs_destroy(cs);
-				cs = NULL;
-				conn = NULL;
-				check->cs = NULL;
 				check->state &= ~CHK_ST_CLOSE_CONN;
+				if (!sc_reset_endp(check->sc)) {
+					/* error will be handled by tcpcheck_main().
+					 * On success, remove all flags except SE_FL_DETACHED
+					 */
+					sc_ep_clr(check->sc, ~SE_FL_DETACHED);
+				}
 				tcpcheck_main(check);
 			}
 			if (check->result == CHK_RES_UNKNOWN) {
@@ -1170,7 +1262,10 @@ struct task *process_chk_conn(struct task *t, void *context, unsigned int state)
 	/* check complete or aborted */
 	TRACE_STATE("health-check complete or aborted", CHK_EV_TASK_WAKE|CHK_EV_HCHK_END, check);
 
+	/* check->sc may be NULL when the healthcheck is purged */
 	check->current_step = NULL;
+	sc = check->sc;
+	conn = (sc ? sc_conn(sc) : NULL);
 
 	if (conn && conn->xprt) {
 		/* The check was aborted and the connection was not yet closed.
@@ -1178,20 +1273,12 @@ struct task *process_chk_conn(struct task *t, void *context, unsigned int state)
 		 * as a failed response coupled with "observe layer7" caused the
 		 * server state to be suddenly changed.
 		 */
-		cs_drain_and_close(cs);
+		sc_conn_drain_and_shut(sc);
 	}
 
-	if (cs) {
-		if (check->wait_list.events)
-			cs->conn->mux->unsubscribe(cs, check->wait_list.events, &check->wait_list);
-		/* We may have been scheduled to run, and the
-		 * I/O handler expects to have a cs, so remove
-		 * the tasklet
-		 */
-		tasklet_remove_from_tasklet_list(check->wait_list.tasklet);
-		cs_destroy(cs);
-		cs = check->cs = NULL;
-		conn = NULL;
+	if (sc) {
+		sc_destroy(sc);
+		check->sc = NULL;
 	}
 
 	if (check->sess != NULL) {
@@ -1200,6 +1287,7 @@ struct task *process_chk_conn(struct task *t, void *context, unsigned int state)
 		check->sess = NULL;
 	}
 
+  end:
 	if (check->server && likely(!(check->state & CHK_ST_PURGE))) {
 		if (check->result == CHK_RES_FAILED) {
 			/* a failure or timeout detected */
@@ -1217,23 +1305,37 @@ struct task *process_chk_conn(struct task *t, void *context, unsigned int state)
 			check_notify_success(check);
 		}
 	}
-	task_set_affinity(t, MAX_THREADS_MASK);
+
+        if (LIST_INLIST(&check->buf_wait.list))
+                LIST_DEL_INIT(&check->buf_wait.list);
+
 	check_release_buf(check, &check->bi);
 	check_release_buf(check, &check->bo);
 	check->state &= ~(CHK_ST_INPROGRESS|CHK_ST_IN_ALLOC|CHK_ST_OUT_ALLOC);
+	check->state |= CHK_ST_SLEEPING;
 
+ update_timer:
 	if (check->server) {
 		rv = 0;
 		if (global.spread_checks > 0) {
 			rv = srv_getinter(check) * global.spread_checks / 100;
-			rv -= (int) (2 * rv * (ha_random32() / 4294967295.0));
+			rv -= (int) (2 * rv * (statistical_prng() / 4294967295.0));
 		}
 		t->expire = tick_add(now_ms, MS_TO_TICKS(srv_getinter(check) + rv));
+		/* reset fastinter flag (if set) so that srv_getinter()
+		 * only returns fastinter if server health is degraded
+		 */
+		check->state &= ~CHK_ST_FASTINTER;
 	}
 
  reschedule:
-	while (tick_is_expired(t->expire, now_ms))
-		t->expire = tick_add(t->expire, MS_TO_TICKS(check->inter));
+	if (proxy->flags & (PR_FL_DISABLED|PR_FL_STOPPED))
+		t->expire = TICK_ETERNITY;
+	else {
+		while (tick_is_expired(t->expire, now_ms))
+			t->expire = tick_add(t->expire, MS_TO_TICKS(check->inter));
+	}
+
  out_unlock:
 	if (check->server)
 		HA_SPIN_UNLOCK(SERVER_LOCK, &check->server->lock);
@@ -1269,16 +1371,18 @@ int check_buf_available(void *target)
 {
 	struct check *check = target;
 
+	BUG_ON(!check->sc);
+
 	if ((check->state & CHK_ST_IN_ALLOC) && b_alloc(&check->bi)) {
 		TRACE_STATE("unblocking check, input buffer allocated", CHK_EV_TCPCHK_EXP|CHK_EV_RX_BLK, check);
 		check->state &= ~CHK_ST_IN_ALLOC;
-		tasklet_wakeup(check->wait_list.tasklet);
+		tasklet_wakeup(check->sc->wait_event.tasklet);
 		return 1;
 	}
 	if ((check->state & CHK_ST_OUT_ALLOC) && b_alloc(&check->bo)) {
 		TRACE_STATE("unblocking check, output buffer allocated", CHK_EV_TCPCHK_SND|CHK_EV_TX_BLK, check);
 		check->state &= ~CHK_ST_OUT_ALLOC;
-		tasklet_wakeup(check->wait_list.tasklet);
+		tasklet_wakeup(check->sc->wait_event.tasklet);
 		return 1;
 	}
 
@@ -1320,13 +1424,6 @@ const char *init_check(struct check *check, int type)
 	check->bi = BUF_NULL;
 	check->bo = BUF_NULL;
 	LIST_INIT(&check->buf_wait.list);
-
-	check->wait_list.tasklet = tasklet_new();
-	if (!check->wait_list.tasklet)
-		return "out of memory while allocating check tasklet";
-	check->wait_list.events = 0;
-	check->wait_list.tasklet->process = event_srv_chk_io;
-	check->wait_list.tasklet->context = check;
 	return NULL;
 }
 
@@ -1346,15 +1443,12 @@ void free_check(struct check *check)
 	}
 
 	task_destroy(check->task);
-	if (check->wait_list.tasklet)
-		tasklet_free(check->wait_list.tasklet);
 
 	check_release_buf(check, &check->bi);
 	check_release_buf(check, &check->bo);
-	if (check->cs) {
-		ha_free(&check->cs->conn);
-		cs_free(check->cs);
-		check->cs = NULL;
+	if (check->sc) {
+		sc_destroy(check->sc);
+		check->sc = NULL;
 	}
 }
 
@@ -1395,11 +1489,8 @@ int start_check_task(struct check *check, int mininter,
 	else
 		t = task_new_anywhere();
 
-	if (!t) {
-		ha_alert("Starting [%s:%s] check: out of memory.\n",
-			 check->server->proxy->id, check->server->id);
-		return 0;
-	}
+	if (!t)
+		goto fail_alloc_task;
 
 	check->task = t;
 	t->process = process_chk;
@@ -1408,15 +1499,28 @@ int start_check_task(struct check *check, int mininter,
 	if (mininter < srv_getinter(check))
 		mininter = srv_getinter(check);
 
+	if (global.spread_checks > 0) {
+		int rnd;
+
+		rnd  = srv_getinter(check) * global.spread_checks / 100;
+		rnd -= (int) (2 * rnd * (ha_random32() / 4294967295.0));
+		mininter += rnd;
+	}
+
 	if (global.max_spread_checks && mininter > global.max_spread_checks)
 		mininter = global.max_spread_checks;
 
 	/* check this every ms */
 	t->expire = tick_add(now_ms, MS_TO_TICKS(mininter * srvpos / nbcheck));
-	check->start = now;
+	check->start = now_ns;
 	task_queue(t);
 
 	return 1;
+
+  fail_alloc_task:
+	ha_alert("Starting [%s:%s] check: out of memory.\n",
+		 check->server->proxy->id, check->server->id);
+	return 0;
 }
 
 /*
@@ -1559,22 +1663,33 @@ int init_srv_check(struct server *srv)
 
 	check_type = srv->check.tcpcheck_rules->flags & TCPCHK_RULES_PROTO_CHK;
 
-	/* If neither a port nor an addr was specified and no check transport
-	 * layer is forced, then the transport layer used by the checks is the
-	 * same as for the production traffic. Otherwise we use raw_sock by
-	 * default, unless one is specified.
-	 */
-	if (!srv->check.port && !is_addr(&srv->check.addr)) {
-		if (!srv->check.use_ssl && srv->use_ssl != -1) {
-			srv->check.use_ssl = srv->use_ssl;
-			srv->check.xprt    = srv->xprt;
+	if (!(srv->flags & SRV_F_DYNAMIC)) {
+		/* If neither a port nor an addr was specified and no check
+		 * transport layer is forced, then the transport layer used by
+		 * the checks is the same as for the production traffic.
+		 * Otherwise we use raw_sock by default, unless one is
+		 * specified.
+		 */
+		if (!srv->check.port && !is_addr(&srv->check.addr)) {
+			if (!srv->check.use_ssl && srv->use_ssl != -1) {
+				srv->check.use_ssl = srv->use_ssl;
+				srv->check.xprt    = srv->xprt;
+			}
+			else if (srv->check.use_ssl == 1)
+				srv->check.xprt = xprt_get(XPRT_SSL);
+			srv->check.send_proxy |= (srv->pp_opts);
 		}
 		else if (srv->check.use_ssl == 1)
 			srv->check.xprt = xprt_get(XPRT_SSL);
-		srv->check.send_proxy |= (srv->pp_opts);
 	}
-	else if (srv->check.use_ssl == 1)
-		srv->check.xprt = xprt_get(XPRT_SSL);
+	else {
+		/* For dynamic servers, check-ssl and check-send-proxy must be
+		 * explicitly defined even if the check port was not
+		 * overridden.
+		 */
+		if (srv->check.use_ssl == 1)
+			srv->check.xprt = xprt_get(XPRT_SSL);
+	}
 
 	/* Inherit the mux protocol from the server if not already defined for
 	 * the check
@@ -1683,6 +1798,13 @@ int init_srv_agent_check(struct server *srv)
 		LIST_INSERT(srv->agent.tcpcheck_rules->list, &chk->list);
 	}
 
+	/* <chk> is always defined here and it is a CONNECT action. If there is
+	 * a preset variable, it means there is an agent string defined and data
+	 * will be sent after the connect.
+	 */
+	if (!LIST_ISEMPTY(&srv->agent.tcpcheck_rules->preset_vars))
+		chk->connect.options |= TCPCHK_OPT_HAS_DATA;
+
 
 	err = init_check(&srv->agent, PR_O2_TCPCHK_CHK);
 	if (err) {
@@ -1711,8 +1833,11 @@ int init_srv_agent_check(struct server *srv)
 
 static void deinit_srv_check(struct server *srv)
 {
-	if (srv->check.state & CHK_ST_CONFIGURED)
+	if (srv->check.state & CHK_ST_CONFIGURED) {
 		free_check(&srv->check);
+		/* it is safe to drop now since the main server reference is still held by the proxy */
+		srv_drop(srv);
+	}
 	srv->check.state &= ~CHK_ST_CONFIGURED & ~CHK_ST_ENABLED;
 	srv->do_check = 0;
 }
@@ -1720,8 +1845,11 @@ static void deinit_srv_check(struct server *srv)
 
 static void deinit_srv_agent_check(struct server *srv)
 {
-	if (srv->agent.state & CHK_ST_CONFIGURED)
+	if (srv->agent.state & CHK_ST_CONFIGURED) {
 		free_check(&srv->agent);
+		/* it is safe to drop now since the main server reference is still held by the proxy */
+		srv_drop(srv);
+	}
 
 	srv->agent.state &= ~CHK_ST_CONFIGURED & ~CHK_ST_ENABLED & ~CHK_ST_AGENT;
 	srv->do_agent = 0;
@@ -1983,20 +2111,6 @@ int set_srv_agent_send(struct server *srv, const char *send)
 	return 0;
 }
 
-/* set agent addr and appropriate flag */
-inline void set_srv_agent_addr(struct server *srv, struct sockaddr_storage *sk)
-{
-	srv->agent.addr = *sk;
-	srv->flags |= SRV_F_AGENTADDR;
-}
-
-/* set agent port and appropriate flag */
-inline void set_srv_agent_port(struct server *srv, int port)
-{
-	srv->agent.port = port;
-	srv->flags |= SRV_F_AGENTPORT;
-}
-
 /* Parse the "agent-send" server keyword */
 static int srv_parse_agent_send(char **args, int *cur_arg, struct proxy *curpx, struct server *srv,
 				char **errmsg)
